#!/usr/bin/env python3
"""
Network Traffic Analyzer
Real-time traffic analysis and anomaly detection.
Parameters:
    - interface: Network interface to monitor
    - mode: Analysis mode (basic, deep, forensic)
Requires: admin/root privileges and scapy
"""
from scapy.all import *
import logging
from datetime import datetime
import json
from pathlib import Path
import threading
import queue
import time
import statistics
import collections
import ipaddress

class TrafficAnalyzer:
    def __init__(self, interface, mode='basic', pcap_file=None):
        self.interface = interface
        self.mode = mode
        self.pcap_file = pcap_file
        
        self.packet_queue = queue.Queue()
        self.stats_queue = queue.Queue()
        self.alert_queue = queue.Queue()
        self.stop_event = threading.Event()
        
        # Traffic statistics
        self.stats = {
            'total_packets': 0,
            'protocols': collections.Counter(),
            'ports': collections.Counter(),
            'ip_sources': collections.Counter(),
            'ip_destinations': collections.Counter(),
            'packet_sizes': [],
            'intervals': [],
            'last_packet_time': None
        }
        
        # Baseline metrics
        self.baselines = {
            'avg_packet_size': None,
            'avg_interval': None,
            'std_dev_size': None,
            'std_dev_interval': None
        }
        
        self.setup_logging()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("traffic_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "traffic.log"),
                logging.StreamHandler()
            ]
        )
        
    def packet_callback(self, packet):
        """Process captured packet"""
        if self.stop_event.is_set():
            return
            
        try:
            # Put packet in queue for analysis
            self.packet_queue.put(packet)
            
            # Basic packet statistics
            self.stats['total_packets'] += 1
            
            # Calculate packet interval
            current_time = time.time()
            if self.stats['last_packet_time']:
                interval = current_time - self.stats['last_packet_time']
                self.stats['intervals'].append(interval)
            self.stats['last_packet_time'] = current_time
            
            # Extract protocol information
            if IP in packet:
                self.stats['ip_sources'][packet[IP].src] += 1
                self.stats['ip_destinations'][packet[IP].dst] += 1
                self.stats['packet_sizes'].append(len(packet))
                
                # Protocol analysis
                if TCP in packet:
                    self.stats['protocols']['TCP'] += 1
                    self.stats['ports'][f"TCP_{packet[TCP].dport}"] += 1
                elif UDP in packet:
                    self.stats['protocols']['UDP'] += 1
                    self.stats['ports'][f"UDP_{packet[UDP].dport}"] += 1
                elif ICMP in packet:
                    self.stats['protocols']['ICMP'] += 1
                    
            # Update baselines periodically
            if len(self.stats['packet_sizes']) >= 1000:
                self.update_baselines()
                
        except Exception as e:
            logging.error(f"Error processing packet: {str(e)}")
            
    def update_baselines(self):
        """Update traffic baselines"""
        try:
            self.baselines.update({
                'avg_packet_size': statistics.mean(self.stats['packet_sizes']),
                'avg_interval': statistics.mean(self.stats['intervals']),
                'std_dev_size': statistics.stdev(self.stats['packet_sizes']),
                'std_dev_interval': statistics.stdev(self.stats['intervals'])
            })
        except Exception as e:
            logging.error(f"Error updating baselines: {str(e)}")
            
    def detect_anomalies(self, packet):
        """Detect traffic anomalies"""
        if not all(self.baselines.values()):
            return
            
        alerts = []
        
        try:
            if IP in packet:
                # Check packet size anomaly
                size = len(packet)
                if abs(size - self.baselines['avg_packet_size']) > (3 * self.baselines['std_dev_size']):
                    alerts.append({
                        'type': 'size_anomaly',
                        'size': size,
                        'avg_size': self.baselines['avg_packet_size']
                    })
                    
                # Check for potential port scan
                if TCP in packet or UDP in packet:
                    proto = 'TCP' if TCP in packet else 'UDP'
                    port = packet[proto].dport
                    if self.stats['ports'][f"{proto}_{port}"] == 1:
                        source = packet[IP].src
                        if self.stats['ip_sources'][source] > 100:
                            alerts.append({
                                'type': 'port_scan',
                                'source': source,
                                'protocol': proto
                            })
                            
                # Check for potential DDoS
                if len(self.stats['intervals']) > 100:
                    current_rate = 1 / statistics.mean(self.stats['intervals'][-100:])
                    if current_rate > 1000:  # More than 1000 packets per second
                        alerts.append({
                            'type': 'high_traffic',
                            'rate': current_rate,
                            'threshold': 1000
                        })
                        
        except Exception as e:
            logging.error(f"Error detecting anomalies: {str(e)}")
            
        return alerts
        
    def analyze_protocols(self, packet):
        """Detailed protocol analysis"""
        analysis = {}
        
        try:
            if IP in packet:
                analysis['ip'] = {
                    'version': packet[IP].version,
                    'src': packet[IP].src,
                    'dst': packet[IP].dst,
                    'ttl': packet[IP].ttl
                }
                
                if TCP in packet:
                    analysis['tcp'] = {
                        'sport': packet[TCP].sport,
                        'dport': packet[TCP].dport,
                        'flags': packet[TCP].flags,
                        'seq': packet[TCP].seq
                    }
                    
                    # HTTP analysis
                    if packet[TCP].dport == 80 or packet[TCP].sport == 80:
                        try:
                            if Raw in packet:
                                http = packet[Raw].load.decode()
                                if http.startswith(('GET', 'POST', 'HTTP')):
                                    analysis['http'] = http.split('\r\n')[0]
                        except:
                            pass
                            
                elif UDP in packet:
                    analysis['udp'] = {
                        'sport': packet[UDP].sport,
                        'dport': packet[UDP].dport,
                        'len': packet[UDP].len
                    }
                    
                    # DNS analysis
                    if packet[UDP].dport == 53 or packet[UDP].sport == 53:
                        if DNS in packet:
                            analysis['dns'] = {
                                'qr': packet[DNS].qr,
                                'opcode': packet[DNS].opcode
                            }
                            
                elif ICMP in packet:
                    analysis['icmp'] = {
                        'type': packet[ICMP].type,
                        'code': packet[ICMP].code
                    }
                    
        except Exception as e:
            logging.error(f"Error analyzing protocols: {str(e)}")
            
        return analysis
        
    def start_capture(self):
        """Start packet capture"""
        logging.info(f"Starting traffic analysis on interface {self.interface}")
        logging.info(f"Mode: {self.mode}")
        
        # Start packet processing thread
        processor = threading.Thread(target=self.process_packets)
        processor.start()
        
        try:
            if self.pcap_file:
                # Read from PCAP file
                sniff(offline=self.pcap_file, prn=self.packet_callback)
            else:
                # Live capture
                sniff(iface=self.interface, prn=self.packet_callback,
                    stop_filter=lambda x: self.stop_event.is_set())
                
        except KeyboardInterrupt:
            logging.info("Stopping capture...")
        finally:
            self.stop_event.set()
            processor.join()
            
            # Save final results
            self.save_results()
            
    def process_packets(self):
        """Process captured packets"""
        while not self.stop_event.is_set() or not self.packet_queue.empty():
            try:
                packet = self.packet_queue.get(timeout=1)
                
                # Detect anomalies
                alerts = self.detect_anomalies(packet)
                if alerts:
                    for alert in alerts:
                        self.alert_queue.put({
                            'timestamp': datetime.now().isoformat(),
                            'alert': alert
                        })
                        
                # Detailed protocol analysis in deep or forensic mode
                if self.mode in ['deep', 'forensic']:
                    analysis = self.analyze_protocols(packet)
                    if analysis:
                        self.stats_queue.put({
                            'timestamp': datetime.now().isoformat(),
                            'analysis': analysis
                        })
                        
            except queue.Empty:
                continue
                
    def save_results(self):
        """Save analysis results"""
        results = {
            'timestamp': datetime.now().isoformat(),
            'interface': self.interface,
            'mode': self.mode,
            'duration': time.time() - self.stats['last_packet_time'],
            'statistics': {
                'total_packets': self.stats['total_packets'],
                'protocols': dict(self.stats['protocols']),
                'top_sources': dict(self.stats['ip_sources'].most_common(10)),
                'top_destinations': dict(self.stats['ip_destinations'].most_common(10)),
                'top_ports': dict(self.stats['ports'].most_common(10))
            }
        }
        
        # Add baselines if available
        if all(self.baselines.values()):
            results['baselines'] = self.baselines
            
        # Get alerts
        alerts = []
        while not self.alert_queue.empty():
            alerts.append(self.alert_queue.get())
        results['alerts'] = alerts
        
        # Get detailed analysis for deep/forensic mode
        if self.mode in ['deep', 'forensic']:
            analysis = []
            while not self.stats_queue.empty():
                analysis.append(self.stats_queue.get())
            results['detailed_analysis'] = analysis
            
        # Save to file
        output_dir = Path("traffic_logs")
        output_file = output_dir / f"analysis_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(output_file, 'w') as f:
            json.dump(results, f, indent=2)
            
        logging.info(f"Analysis results saved to {output_file}")
        
        # Print summary
        print("\nTraffic Analysis Summary")
        print("=" * 50)
        print(f"Total Packets: {results['statistics']['total_packets']}")
        print("\nProtocol Distribution:")
        for proto, count in results['statistics']['protocols'].items():
            print(f"{proto}: {count}")
        print("\nTop Sources:")
        for ip, count in results['statistics']['top_sources'].items():
            print(f"{ip}: {count} packets")
        print(f"\nTotal Alerts: {len(alerts)}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Network Traffic Analyzer")
    parser.add_argument("interface", help="Network interface to monitor")
    parser.add_argument("--mode", choices=['basic', 'deep', 'forensic'],
                      default='basic', help="Analysis mode")
    parser.add_argument("--pcap", help="PCAP file to analyze")
    args = parser.parse_args()
    
    analyzer = TrafficAnalyzer(args.interface, args.mode, args.pcap)
    analyzer.start_capture()
